@@ -1,6 +1,6 @@
 /*
- * marsh:  a short program to read and save data
- *         from a MAS-345 multimeter to a file.
+ * marsh:  a short program to read and save data from
+ *         MAS-345 and Digitek-4000 multimeters to a file.
  *
  * Copyright (C) 2005 Marcello Carla'
  *
@@ -32,7 +32,7 @@
  *
  */
 
-#include <stdio.h>
+#define _GNU_SOURCE
 #include <fcntl.h>
 #include <errno.h>
 #include <termios.h>
@@ -44,29 +44,40 @@
 #include <time.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <stdio.h>
 
-#define  DATA_STRING 14   /* length of data string from MAS-345 */
-#define  MAX_PORTS 4      /* can be increased up to 32 */
-#define  MAX_TIME_OUT 7   /* calculation disabled after these many failures */
+#define MAX_PORTS 4                  /* can be increased up to 32 */
+#define DEFAULT_DEVICE   mas345
+#define DEFAULT_MODEL   "MAS-345"
+#define DEFAULT_PORT   "/dev/ttyS0"
+#define DEFAULT_TIMEOUT 2.0          /* time to complete measurements (sec) */
+
+#define  MAX_TIME_OUT 7   /* disable calculations after these many failures */
+#define  DATA_STRING 15   /* max length of input data string */
 #define  MAX_RESULT 40    /* max length for result string */
 
 #define  DTR 2
 #define  RTS 4
 
-  struct port_area {
-    char *port;                 /* device name */
-    int  fd;                    /* file descriptor */
-    struct termios oldtio;      /* port setting as found at start */
-    char inbuf[DATA_STRING+1];  /* data area for input string */
-    int  j;                     /* character count in inbuf */
-    int  mask;                  /* flag position in mask */
-  };
+struct port_area {
+  char *port;                 /* port name */
+  char *model;                /* device name */
+  void (**actions)();         /* device specific routines */
+  int  fd;                    /* file descriptor */
+  struct termios oldtio;      /* port setting as found at start */
+  unsigned char
+       inbuf[DATA_STRING+1];  /* data area for input string */
+  int  j;                     /* character count in inbuf */
+  int  mask;                  /* flag position in mask */
+  int  position,length;       /* position and length of datum in inbuf */
+  struct timeval ta,tb;       /* time structures for diagnostics */
+};
 
 struct termios newtio;
 int  rts = RTS, dtr = DTR;
-char *default_port = "/dev/ttyS0";
-char *default_out = "mas345.dat";
 int D = 0, DD = 0, DDD = 0;                 /* verbose debug flag */
+char *default_model = DEFAULT_MODEL;
+char *default_port = DEFAULT_PORT;
 
 /*
  * restore_port_settings - to be called on exit for clean operation
@@ -100,6 +111,7 @@ void catch_the_signal (signal) {
     return;
   }
   if (D) printf ("D - received signal %d\n", signal);
+  fcloseall();         /* close all streams, mainly the data to file */
   exit (0);
 }
 
@@ -118,6 +130,8 @@ void close_calculator (int exit_value, int fd_elab[]) {
  * time operations: difference and sum
  */
 
+/* return time difference as a timeval structure */
+
 void time_diff (struct timeval *a, struct timeval *b, struct timeval *d) {
   d->tv_sec = a->tv_sec - b->tv_sec;
   d->tv_usec = a->tv_usec - b->tv_usec;
@@ -127,6 +141,8 @@ void time_diff (struct timeval *a, struct timeval *b, struct timeval *d) {
   }
 }
 
+/* return time sum as a timeval structure */
+
 void time_add (struct timeval *a, struct timeval *b, struct timeval *s) {
   s->tv_sec = a->tv_sec + b->tv_sec;
   s->tv_usec = a->tv_usec + b->tv_usec;
@@ -136,14 +152,28 @@ void time_add (struct timeval *a, struct timeval *b, struct timeval *s) {
   }
 }
 
+/* return time difference in millisec as a float quantity */
+
+float delta_time (struct timeval *a, struct timeval *b) {
+  return ((a->tv_sec - b->tv_sec)*1000. + 
+	  (a->tv_usec - b->tv_usec)/1000.);
+}
+
+/* convert a float time (in sec) to a timeval structure */
+
+void time_to_timeval (float t, struct timeval *ts) {
+  ts->tv_sec = t;
+  ts->tv_usec = (t - ts->tv_sec)*1e6;
+}
+
 /*
- * time_stamp - convert timeval to a 24 chars string with format:
+ * time_stamp - convert timeval to a char string with format:
  *
- *      yyyy-mm-dd  hh:mm:ss.dd
+ *      yyyy-mm-dd  hh:mm:ss.ddz
  *                1    1    2
  *      0    5    0    5    0
  *
- * and return the string length
+ * and return the string and its length (23 chars + trailing \0)
  *
  */
 
@@ -159,7 +189,6 @@ int time_stamp(char *buf, struct timeval *tv) {
   return (23);
 }
 
-
 /*
  * open and setup the port
  */
@@ -204,13 +233,6 @@ void open_setup_line (struct port_area *p) {
     exit (-1);
   }    
 
-  if (D) printf("D - going to clear RTS and set DTR\n");
-  ioctl (p->fd, TIOCMBIC, &rts);      /* clear RTS line for - supply */
-  ioctl (p->fd, TIOCMBIS, &dtr);      /* set DTR line for + supply */
-
-  usleep(1000);
-  tcflush(p->fd, TCIOFLUSH);
-
   if (D) printf("D - done with line setup\n");
   return;
 }
@@ -219,7 +241,7 @@ void open_setup_line (struct port_area *p) {
  * start the calculator
  */
 
-void start_calc (char **elab, char * elab_p, int fd_elab[]) {
+int start_calc (char * elab_p, int fd_elab[]) {
 
   int pid_comp;
 
@@ -227,8 +249,7 @@ void start_calc (char **elab, char * elab_p, int fd_elab[]) {
 
   if (pipe(fd_elab) || pipe(fd_elab+2)) {
     perror ("### Failed creating pipes for calculator. Calculation disabled");
-    *elab = 0;
-    return;
+    return (-1);
   }
   if (D) printf ("D - created pipes (%d %d),(%d %d)\n",
 		 fd_elab[0], fd_elab[1], fd_elab[2], fd_elab[3]);
@@ -237,8 +258,7 @@ void start_calc (char **elab, char * elab_p, int fd_elab[]) {
     perror ("### 'fork' failed - calculation disabled");
     close (fd_elab[0]); close (fd_elab[1]);
     close (fd_elab[2]); close (fd_elab[3]);
-    *elab = 0;
-    return;
+    return (-1);
   }
 
   if (D) printf ("D - forked. Here pid %d\n", pid_comp);
@@ -279,7 +299,7 @@ void start_calc (char **elab, char * elab_p, int fd_elab[]) {
     write (fd_elab[1], "\n", 1);
   }
 
-  return;
+  return (0);
 }
 
 /*
@@ -297,10 +317,11 @@ int local_elab (char **elab,
   struct timeval t_calc;
   static int timeouts = MAX_TIME_OUT;
 
-  /* send data */
+  /* send data to calculator: m0=<reading 0>;m1=<reading 1>; etc. */
 
   for (j = 0 ; j <= last_port ; j++) {
-    sprintf (buf, "m%d=%6.6s;", j, (p+j)->inbuf+3);
+    sprintf (buf, "m%d=%*.*s;", j, (p+j)->length, (p+j)->length,
+	     (p+j)->inbuf+(p+j)->position);
     if (DDD) printf ("DDD - sending to child: <%s>\n", buf);
     write (fd_elab[1], buf, strlen(buf));
   }
@@ -369,15 +390,249 @@ int local_elab (char **elab,
 }
 
 /*
- * write_to_file
+ * MAS-345
  */
 
-void write_to_file (int fd, char * string) {
-  write (fd, "# ", 2);
-  write (fd, string, strlen(string));
-  write (fd, "\n", 1);
+void mas345_open (struct port_area *p) {
+
+  bzero (&newtio, sizeof(newtio));          /* setup termios structure */
+
+  newtio.c_cflag = CS7 | CSTOPB | CLOCAL | CREAD;
+  newtio.c_cc[VMIN] = 14;
+
+  cfsetospeed (&newtio, B600);
+  cfsetispeed (&newtio, B600);
+
+  open_setup_line (p);
+
+  if (D) printf("D - going to clear RTS and set DTR\n");
+  ioctl (p->fd, TIOCMBIC, &rts);      /* clear RTS line for - supply */
+  ioctl (p->fd, TIOCMBIS, &dtr);      /* set DTR line for + supply */
+
+  usleep(1000);
+  tcflush(p->fd, TCIOFLUSH);
+
+  p->position = 3;     /* postion and length of datum in input buffer */
+  p->length = 6;
 }
 
+void mas345_start (struct port_area *p) {
+  tcflush(p->fd, TCIOFLUSH);
+  write (p->fd, "\n", 1);
+  gettimeofday (&p->ta, NULL);
+}
+
+void mas345_read (struct port_area *p, int *mask) {
+
+#define STRING 14
+
+  int j;
+
+  /* ... check for unwanted characters ...*/
+
+  if (p->mask & ~mask[0]) {
+    tcflush (p->fd, TCIOFLUSH);
+    fprintf (stderr,
+	     "### Spurious characters from device %s on line %s - flushed!\n",
+	     p->model, p->port);
+    return;
+  }
+
+  p->j += read (p->fd, p->inbuf + p->j,
+		  STRING - p->j);        /*  ... and read */
+
+  gettimeofday (&p->tb, NULL);
+
+  if (DDD) printf ("DDD - port %s byte %d: %o\n",
+		   p->port, p->j, *(p->inbuf + p->j - 1));
+  
+  if (p->j == STRING) {      /* if string is complete check for sanity */
+
+    mask[0] &= ~p->mask;           /* let's suppose everything is o.k. */
+    mask[1] &= ~p->mask;
+
+    for (j = 0; j < p->j - 1 ; j++) {                /* check characters? */
+      if (! isprint(p->inbuf[j])) {
+	fprintf (stderr,
+		 "### Bad character from device %s on line %s: %o at %d\n",
+		 p->model, p->port, p->inbuf[j], j);
+	p->inbuf[j] = ' ';
+	mask[1] |= p->mask;                      /* data are not valid */
+      }
+    }
+
+    if (p->inbuf[p->j-1] == '\r') {          /* check terminator? */
+      p->j--;
+    } else {
+      fprintf (stderr,
+	       "### Bad terminator from device %s on line %s: %o\n",
+	       p->model, p->port, p->inbuf[p->j-1]);
+      mask[1] |= p->mask;
+    }
+
+    if (!strncmp(p->inbuf+5, ".OL", 3) ||               /* overflow? */
+	!strncmp(p->inbuf+5, "O.L", 3) ||
+	!strncmp(p->inbuf+5, "OL.", 3)) {
+      fprintf (stderr, "### Overflow in device %s on line %s\n",
+	       p->model, p->port);
+      mask[1] |= p->mask;
+    }
+  }
+}
+
+void (*mas345[])() = {mas345_open, mas345_start, mas345_read};
+
+
+/*
+ * Digitek 4000
+ */
+
+void digitek_open (struct port_area *p) {
+
+  bzero (&newtio, sizeof(newtio));          /* setup termios structure */
+
+  newtio.c_cflag = CS8 | CLOCAL | CREAD;
+  newtio.c_cc[VMIN] = 1;    /* Digitek-4000 is not reliable enough
+			       to allow reading the full string in
+			       a single operation */
+
+  cfsetospeed (&newtio, B2400);
+  cfsetispeed (&newtio, B2400);
+
+  open_setup_line (p);
+
+  if (D) printf("D - going to clear RTS and set DTR\n");
+
+  ioctl (p->fd, TIOCMBIC, &rts);      /* clear RTS line */
+  ioctl (p->fd, TIOCMBIC, &dtr);      /* clear DTR line */
+
+  usleep(1000);
+  tcflush(p->fd, TCIOFLUSH);
+
+  p->position = 0;     /* postion and length of datum in input buffer */
+  p->length = 6;
+}
+
+void digitek_start (struct port_area *p) {
+
+  tcflush(p->fd, TCIOFLUSH);
+  ioctl (p->fd, TIOCMBIS, &dtr);      /* set DTR line to start operation */
+  gettimeofday (&p->ta, NULL);
+
+}
+
+void digitek_read (struct port_area *p, int *mask) {
+
+#define STRING 14
+
+  /* each byte from Digitek-4000 contains a sequence code in upper four bits */
+
+  unsigned char sequence[] = {0020, 0040, 0060, 0100, 0120, 0140, 0160,
+			      0200, 0220, 0240, 0260, 0300, 0320, 0340};
+
+  /* look-up table for 7 segments to bcd conversion */
+
+  char equiv[8][16] = {" ####1##########","#####7#########3",
+		       "#######4########","##############59",
+		       "################","###########2####",
+		       "########L#######","#############068"};
+
+  int j, efg, abcd, dot=0;
+  char value[16] = "               ";
+
+  if (p->mask & ~mask[0]) {
+    tcflush (p->fd, TCIOFLUSH);
+    fprintf (stderr,
+	     "### Spurious characters from device %s on line %s - ignored\n",
+	     p->model, p->port);
+    return;
+  }
+
+  read (p->fd, &p->inbuf[p->j], 1);
+  gettimeofday (&p->tb, NULL);
+
+  if ((p->inbuf[p->j] & 0360) == sequence[p->j]) {
+    p->j++;
+  } else {
+    p->j = 0;
+  }
+
+  if (p->j == STRING) {
+
+    ioctl (p->fd, TIOCMBIC, &dtr);    /* stop transmission */
+
+    mask[0] &= ~p->mask;
+    mask[1] &= ~p->mask;
+
+    /* input string is complete; translate filling buffer "value" */
+
+    /* bytes 1-2, 3-4, 5-6 and 7-8 code for one digit+dot each pair */
+
+    for (j = 0; j < 4 ; j++) {                 /* process digits */
+      if (j && p->inbuf[j+j+1] & 8) {          /* decimal dot here? */
+	value[j+1] = '.';
+	dot = 1;
+      }
+      efg = p->inbuf[j+j+1] & 7;               /* 3 out of 7 segments */
+      abcd = p->inbuf[j+j+2] & 15;             /* 4 out of 7 segments */
+      /* pick character 'abcd' in string 'efg' in table 'equiv' */
+      value[j+1+dot] = equiv[efg][abcd];
+      if (value[j+1+dot] == 'L') {             /* 0L -> OL in overflow */
+        if (value[j+dot] == '0') value[j+dot] = 'O';
+	else value[j+dot-1] = 'O';
+	mask[1] |= p->mask;                    /* data are not valid */
+      }
+    }
+
+    /* dot in first digit is the - sign */
+
+    if (p->inbuf[1] & 8) value[0] = '-';
+
+
+    if (dot == 0) value[5] = '.';              /* no dot -> final dot */
+    j = 1;
+    while (value[j] == '0' && value[j+1] != '.') {  /* remove leading zero's */
+      value[j] = value[j-1];
+      value[j-1] = ' ';
+      j++;
+    }
+
+    if (p->inbuf[12] & 12) {                   /* Volt or Ampere */
+      value[8] = 'V';                          /* maybe Volt */
+      if (p->inbuf[12] & 8) value[8] = 'A';    /* Ampere */
+      if (p->inbuf[10] & 8) value[7] = 'm';    /* milli */
+      if (p->inbuf[9]  & 8) value[7] = 'u';    /* micro */
+      if (p->inbuf[0]  & 4) value[10] = 'D';   /* AC/DC */
+      if (p->inbuf[0]  & 8) value[10] = 'A';
+      value[11] = 'C';
+    } else if (p->inbuf[11] & 4) {             /* ohm */
+      memcpy (value+8, "ohm", 3);
+      if (p->inbuf[10] & 2) value[7] = 'M';    /* Mega  */
+      if (p->inbuf[9]  & 2) value[7] = 'k';    /* kilo  */
+    } else if (p->inbuf[12] & 2) {             /* Hz */
+      value[8] = 'H';
+      value[9] = 'z';
+      if (p->inbuf[10] & 2) value[7] = 'M';    /* Mega  */
+      if (p->inbuf[9]  & 2) value[7] = 'k';    /* kilo  */
+    } else if (p->inbuf[11] & 8) {             /* nF */ 
+      value[8] = 'n';
+      value[9] = 'F';
+    } else if (p->inbuf[13] & 4) {             /* degree */
+      value[8] = 'C';
+    }
+    if (p->inbuf[0] & 2) value[12] = 'a';      /* autorange */
+    else value[12] = 'm';                      /* manual */
+
+    value[13] = '0' + (p->inbuf[11] & 3);      /* 1: Hold  2: Relative */
+
+    *mask &= ~p->mask;
+    memcpy (p->inbuf, value, 15);
+    p->j = 14;
+  }
+}
+
+void (*digitek[])() = {digitek_open, digitek_start, digitek_read};
+
 /*
  *   "the main"
  */
@@ -385,34 +640,41 @@ void write_to_file (int fd, char * string) {
 int main (int argc, char *argv[]) {
 
   struct port_area p[MAX_PORTS];
-  int last_port = 0, mask, mask_prototype = 0;
+  int last_port = 0, mask[2], mask_prototype = 0;
   struct sigaction act;
   int j, l, stamp_s=0, stamp_t=0, file_out=0, file_only=0, sel;
   int out_fd;
+  FILE *out_stream;
   int fd_elab[4]={-1,-1,-1,-1};     /* descriptors for pipe to calculator */
   fd_set rfds;
-  char c, *out_file=default_out, *heading = 0;
+  char c, *out_file, *heading = 0;
   char *elab = 0, *elab_p = 0;
   struct timeval request, tb, ts;         /* data structures for timing */
   struct timeval interval, begin, future, elapsed, now;
+  struct timeval timeout, deadline;
   char outbuf[40 + MAX_RESULT + MAX_PORTS * (DATA_STRING + 1)];
-  float delay = 1.;              /* delay before reading */
-  int repeat = -1, nc;           /* repetition; -1 = endless */
+  float delay = 1., value;
+  int repeat = -1, nc;
   struct itimerval timer;
 
   /*  printf ("### marsh - Copyright (C) 2005 Marcello Carla' ###\n"); */
 
-  p[0].port = default_port;
+  p[0].port = default_port;        /* preset defaults for first device */
+  p[0].actions = DEFAULT_DEVICE;
+  p[0].model = default_model;                 
+
+  time_to_timeval (DEFAULT_TIMEOUT, &timeout);
 
   /* parse command line options */
 
 #define display \
-"\nmarsh - get readings from a MAS345 multimeter via RS232 line\n"\
+"\nmarsh - get readings from multimeters via RS232 line\n"\
 "\n"\
 "Usage: marsh [OPTION]\n"\
 "\n"\
 " -l   (string) line to be used (default: /dev/ttyS0)\n"\
 " -m   (string) one more line for one more multimeter\n"\
+" -j   (string) model for one more multimeter\n"\
 " -i   (float) time interval between readings (sec.)\n"\
 " -r   (int) measure repetition count (default: endless)\n"\
 " -s   add time stamp\n"\
@@ -423,10 +685,11 @@ int main (int argc, char *argv[]) {
 " -e   (string) on line calculations with 'bc'; readings are m0, m1, ... \n"\
 " -p   (string) preset string for external calculator\n"\
 " -d   activate verbose debug (more d's more debug) \n"\
+" -z   (float) time for multimeter to complete \n"\
 " -h   print this help and exit\n"
 
   opterr = 0;
-  while ((c = getopt (argc, argv, "di:l:m:r:hsto::f::c:e:p:")) != -1)
+  while ((c = getopt (argc, argv, "di:l:m:j:r:hsto:f:c:e:p:z:")) != -1)
     switch (c)
       {
       case 'd':
@@ -454,6 +717,21 @@ int main (int argc, char *argv[]) {
 	  exit (-1);
 	}
 	p[last_port].port = optarg;
+        p[last_port].actions = p[last_port-1].actions;
+        p[last_port].model = p[last_port-1].model;
+	break;
+      case 'j':
+	if (optarg[0] == 'm') {
+	  p[last_port].actions = mas345;        /* device is a MAS345 */
+	  p[last_port].model = "MAS-345";
+	} else if (optarg[0] == 'd') {
+	  p[last_port].actions = digitek;       /* device is a Digitek*/
+	  p[last_port].model = "Digitek-4000";
+	} else {
+	  fprintf (stderr, "Model <%s> is unknown\n", optarg);
+	  fprintf (stderr, "Enter m for MAS-345 or d for Digitek\n");
+	  exit (-1);
+	}
 	break;
       case 'r':
 	repeat = atoi(optarg);
@@ -488,6 +766,13 @@ int main (int argc, char *argv[]) {
       case 'p':
 	elab_p = optarg;
 	break;
+      case 'z':
+	if (sscanf(optarg, "%f", &value) != 1) {
+	  fprintf (stderr, "### Bad argument for timeout: <%s>\n", optarg);
+	  exit (-1);
+	}
+	time_to_timeval (value, &timeout);
+	break;
       case 'h':
       case '?':
       default:
@@ -524,10 +809,14 @@ int main (int argc, char *argv[]) {
 
   /* start the calculator, if requested */
 
-  if (elab)
-    start_calc (&elab, elab_p, fd_elab);
-  if (D) printf ("D - started calculator with i/o %d %d\n",
-		 fd_elab[1], fd_elab[2]);
+  if (elab) {
+    if (start_calc (elab_p, fd_elab)) {
+      elab = 0;                        /* failed - disable calculations */
+    } else {
+      if (D) printf ("D - started calculator with i/o %d %d\n",
+		   fd_elab[1], fd_elab[2]);
+    }
+  }
 
   /* build the 'interval' structure from 'delay' */
 
@@ -538,40 +827,39 @@ int main (int argc, char *argv[]) {
 
   /* open and setup the lines */
 
-  bzero (&newtio, sizeof(newtio));          /* setup termios structure */
-
-  newtio.c_cflag = CS7 | CSTOPB | CLOCAL | CREAD;
-  newtio.c_cc[VMIN] = 14;
-
-  cfsetospeed (&newtio, B600);
-  cfsetispeed (&newtio, B600);
-
-
   for (l = 0 ; l <= last_port ; l++ ) {     /* open lines and set param's */
-    open_setup_line (p+l);
+    p[l].actions[0] (&p[l]);
     mask_prototype |= p[l].mask = 1 << l;
   }
 
   /* open output file (if any) */
 
   if (file_out) {
-    if ((out_fd =
-	 open (out_file, O_WRONLY | O_CREAT | O_EXCL, 0644)) == -1) {
+
+    out_fd = open (out_file, O_WRONLY | O_CREAT | O_EXCL, 0644);
+    if (out_fd == -1) {
       fprintf (stderr, "### Unable to open file <%s>: %s\n",
-	      out_file, strerror(errno));
+	       out_file, strerror(errno));
       exit (-1);
     }
-    gettimeofday (&begin,NULL);
+    out_stream = fdopen (out_fd, "w");
 
+    /* write a header to the file */
+
+    gettimeofday (&begin,NULL);                  /* time stamp */
     nc = time_stamp (outbuf, &begin);
-
-    write_to_file (out_fd, outbuf);
-    if (heading) write_to_file (out_fd, heading);
+    fprintf (out_stream, "#\n# %s\n", outbuf);
     
-    for (l = 0 ; l <= last_port ; l++ ) write_to_file (out_fd, p[l].port);
+    if (heading) fprintf (out_stream, "# %s\n", heading);
 
-    if (elab) write_to_file (out_fd, elab);
-    if (elab_p) write_to_file (out_fd, elab_p);
+    for (l = 0 ; l <= last_port ; l++ )             /* devices and lines */
+      fprintf (out_stream, "# <%s> on line <%s>\n",
+	       p[l].model, p[l].port);
+
+    if (elab)   fprintf (out_stream, "# -e: <%s>\n", elab);   /* bc commands */
+    if (elab_p) fprintf (out_stream, "# -p: <%s>\n", elab_p);
+
+    fprintf (out_stream, "#\n");
   }
 
   /* start the time count */
@@ -587,102 +875,73 @@ int main (int argc, char *argv[]) {
 
   while (repeat>0 ? repeat-- : repeat) {
 
-    /* send a solicit character to each active line and clear the buffers */
+    /* send a request to each active line and clear the buffers */
 
-    mask = mask_prototype;
+    mask[0] = mask[1] = mask_prototype;
     gettimeofday (&request, NULL);
+    time_add (&request, &timeout, &deadline);
+
     for (l = 0 ; l <= last_port ; l++ ) {
-      write (p[l].fd, "\n", 1);
       p[l].inbuf[0] = '\0';
       p[l].j = 0;
+      p[l].actions[1] (&p[l]);
     }
-    if (DD) printf ("DD - count %d   mask = %o\n", repeat, mask);
 
-    while (mask) {
+    if (DD) printf ("DD - count %d   mask = %o\n", repeat, mask[0]);
+
+    /* receive from each line or timeout */
+
+    while (mask[0]) {
 
       FD_ZERO (&rfds);             /* prepare the select() data structure */
       for (l = 0 ; l <= last_port ; l++ ) FD_SET (p[l].fd, &rfds);
-      ts.tv_sec = 1;
-      ts.tv_usec = 0;
+
+      gettimeofday (&now, NULL);
+      time_diff (&deadline, &now, &ts);
+      if (ts.tv_sec < 0) {
+	ts.tv_sec = 0;
+	ts.tv_usec = 0;
+      }
 
       if ((sel = select (p[last_port].fd+1, &rfds, NULL, NULL, &ts)) > 0) {
 
-	for (l = 0 ; l <= last_port ; l++ ) {         /* for each port ... */
+	/* for each port, if ready, call read action to manage input */
 
-	  if (FD_ISSET(p[l].fd, &rfds)) {             /* ... if ready ...  */
+	for (l = 0 ; l <= last_port ; l++ ) {
 
-                                   /* ... check for unwanted characters ...*/
-	    if (p[l].mask & ~mask) {
-	      tcflush (p[l].fd, TCIOFLUSH);
-	      fprintf (stderr,
-		       "### Spurious characters on line %d - flushed!\n", l);
-	    }
+	  if (FD_ISSET(p[l].fd, &rfds)) {
 
-	    p[l].j += read (p[l].fd, p[l].inbuf + p[l].j,
-			    DATA_STRING - p[l].j);        /*  ... and read */
-	    if (DDD) printf ("DDD - port %d byte %d: %o\n",
-			     l, p[l].j, *(p[l].inbuf + p[l].j - 1));
+	    p[l].actions[2] (&p[l], mask);
 
-	    if (p[l].j == DATA_STRING) {      /* if string is complete ... */
-
-                                          /* ... and properly terminated ... */
-	      if (*(p[l].inbuf + p[l].j - 1) == '\r') {
-		mask &= ~p[l].mask;                        /* ... close it */
-		*(p[l].inbuf + p[l].j - 1) = ' ';
-
-	      } else {            /* otherwise complain and flush the line */
-
-		fprintf (stderr,
-			 "### Error on input %d: <%s>\n", l, p[l].inbuf);
-		tcflush (p[l].fd, TCIOFLUSH);
-	      }
-	    }
 	  }
 	}
 
       } else if (sel == 0) {
 	fprintf (stderr, "### Time out on input on:\n");
 	for (l = 0 ; l <= last_port ; l++ ) {
-	  if (mask & p[l].mask) {
-	    *(p[l].inbuf + p[l].j) = '\0';
-	    fprintf (stderr, "   port %d after %d bytes\n", l, p[l].j);
-	    tcflush (p[l].fd, TCIOFLUSH);
+	  if (mask[0] & p[l].mask) {
+	    fprintf (stderr, "model %s on line %s after %d bytes\n",
+		     p[l].model, p[l].port, p[l].j);
 	  }
 	}
-	break;
+	break;          /* time is over - exit from the 'while mask' loop */
       } else {
 	perror ("### Unable to select()");
 	exit (-1);
       }
     }
 
-    for (l = 0 ; l <= last_port ; l++ ) {
-
-      for (j = 0; j < p[l].j ; j++) {      /* sanity check for result string */
-	if (! isprint(p[l].inbuf[j])) {
-	  fprintf (stderr, "### Bad character from calculator: %o at %d\n",
-		   p[l].inbuf[j], j);
-	  p[l].inbuf[j] = ' ';
-	  mask |= ~p[l].mask;
-	}
-      }
-      if (!strncmp(p[l].inbuf+5, ".OL", 3) ||
-	  !strncmp(p[l].inbuf+5, "O.L", 3) ||
-	  !strncmp(p[l].inbuf+5, "OL.", 3)) {
-	fprintf (stderr, "### Measurement overflow!\n");
-	mask |= ~p[l].mask;
-      }
-      if (DD) {
-	printf ("DD - line: %d   mode:<%2.2s>  value:<%6.6s>  unit:<%4.4s>\n", 
-		l, p[l].inbuf, p[l].inbuf+3, p[l].inbuf+9);
+    if (DD) {
+      for (l = 0 ; l <= last_port ; l++ ) {
+	printf ("DD - line: %d: <%s>\n", l, p[l].inbuf);
       }
     }
 
-    /* write time section to output bufer */
+    /* write time section to output buffer */
 
     gettimeofday (&tb,NULL);
 
-    if (mask) sprintf (outbuf, "#");     /* there is some error */
+    if (mask[1]) sprintf (outbuf, "#");     /* data are not valid */
     else sprintf (outbuf, " ");
     nc = 1;
 
@@ -702,7 +961,7 @@ int main (int argc, char *argv[]) {
 
       p[l].inbuf[p[l].j] = '\0';
       if (DD) printf("DD - line %d: <%s>\n", l, p[l].inbuf);
-      if ((mask & p[l].mask) == 0) {
+      if ((mask[0] & p[l].mask) == 0) {
 	nc += sprintf (outbuf+nc, "   %s", p[l].inbuf);
       } else
 	nc += sprintf (outbuf+nc, "                 ");
@@ -711,7 +970,7 @@ int main (int argc, char *argv[]) {
     /* execute on line data elaboration */
 
     if (DD) printf ("DD - starting optional elaboration\n");
-    if (elab && !mask) {
+    if (elab && !mask[1]) {
       outbuf[nc++] = ' '; outbuf[nc++] = ' ';
       nc += l = local_elab (&elab, p, last_port, outbuf+nc, fd_elab);
       if (l == 0 && elab) outbuf[0] = '#';    /* mark for bad record */
@@ -721,7 +980,7 @@ int main (int argc, char *argv[]) {
     /* write output buffer to console & file */
 
     if (!file_only) printf ("%s", outbuf);
-    if (file_out) write (out_fd, outbuf, nc);
+    if (file_out) fprintf (out_stream, "%s", outbuf);
 
     /* schedule next operation */
 
@@ -748,6 +1007,8 @@ int main (int argc, char *argv[]) {
       }
     }
 
+/* now set the timer and go to sleep until next alarm */
+
     if (setitimer(ITIMER_REAL, &timer, NULL)) {
       perror ("### Couldn't start timer");
       exit (-1);
